//==================================================================
//
//	File name	:	jaswin.js
//
//	Summary		:	Jaswin common functions and classes
//
//	Ceated		:	2018/11/15
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

var global;		// Global instance

(function (jaswin) {
	//==========================================================
	//	Global
	//==========================================================
	// Browser
	jaswin.BROWSER = "";
	
	// Error type
	jaswin.ERR_ALERT = 1;
	jaswin.ERR_EXCEPTION = -1;
	
	// Align
	jaswin.ALIGN_LEFT = 0;
	jaswin.ALIGN_RIGHT = 1;
	jaswin.ALIGN_CENTER = 2;
	
	// Anchor
	jaswin.ANCHOR_BOTH = 0;
	jaswin.ANCHOR_LEFT = 1;
	jaswin.ANCHOR_RIGHT = 2;
	jaswin.ANCHOR_TOP = 1;
	jaswin.ANCHOR_BOTTOM = 2;
	
	// Drop type
	jaswin.DROP_TEXT = 0;
	jaswin.DROP_URL = 1;
	jaswin.DROP_HTML = 2;
	jaswin.DROP_FILE = 3;
	
	// Icons
	jaswin.ICON_TRIANGLE_TOP				= "triangle-top";
	jaswin.ICON_TRIANGLE_BOTTOM				= "triangle-bottom";
	jaswin.ICON_TRIANGLE_LEFT				= "triangle-left";
	jaswin.ICON_TRIANGLE_RIGHT				= "triangle-right";
	jaswin.ICON_TRIANGLE_TOP_DISABLED		= "triangle-top-disabled";
	jaswin.ICON_TRIANGLE_BOTTOM_DISABLED	= "triangle-bottom-disabled";
	jaswin.ICON_TRIANGLE_LEFT_DISABLED		= "triangle-left-disabled";
	jaswin.ICON_TRIANGLE_RIGHT_DISABLED		= "triangle-right-disabled";
	jaswin.ICON_INFO						= "icon-info";
	jaswin.ICON_QUESTION					= "icon-question";
	jaswin.ICON_ALERT						= "icon-alert";
	jaswin.ICON_ERROR						= "icon-error";
	jaswin.ICON_FOLDER						= "icon-folder";
	jaswin.ICON_FILE						= "icon-file";
	jaswin.ICON_OPEN						= "icon-open";
	jaswin.ICON_SAVE						= "icon-save";
	jaswin.ICON_CHECKBOX					= "icon-check";
	jaswin.ICON_CHECKBOX_HOVER				= "icon-check-hover";
	jaswin.ICON_CHECKBOX_CHECKED			= "icon-check-checked";
	jaswin.ICON_CHECKBOX_CHECKED_DISABLED	= "icon-check-checked-disabled";
	jaswin.ICON_RADIO						= "icon-radio";
	jaswin.ICON_RADIO_HOVER					= "icon-radio-hover";
	jaswin.ICON_RADIO_CHECKED				= "icon-radio-checked";
	jaswin.ICON_RADIO_CHECKED_DISABLED		= "icon-radio-checked-disabled";
	
	// Message box
	jaswin.MSGBOX_OK						= 0;
	jaswin.MSGBOX_INFO						= 1;
	jaswin.MSGBOX_QUESTION					= 2;
	jaswin.MSGBOX_ALERT						= 4;
	jaswin.MSGBOX_ERROR						= 8;
	jaswin.MSGBOX_OKCANCEL					= 16;
	jaswin.MSGBOX_YESNO						= 32;
	jaswin.MSGBOX_RETURN_CANCEL				= 0;
	jaswin.MSGBOX_RETURN_OKYES				= 1;
	
	// Global size
	jaswin.currentWidth = 0;
	jaswin.currentHeight = 0;
	jaswin.lastWidth = 0;
	jaswin.lastHeight = 0;
	
	// Current window
	jaswin.currentWindow = null;
	
	
	//==========================================================
	//
	//	Class		:	JaswinError
	//
	//	Summary		:	Error throwed by jaswin
	//
	//==========================================================
	jaswin.JaswinError = class extends Error
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iType,			// Error type[int]
					sMessage,		// Error message[string]
					sFilename,		// File name[string]
					sfunctionName)	// Function name[string]
		{
			super(sMessage);
			this.type = iType;
			this.message = sMessage;
			this.fileName = sFilename;
			this.functionName = sfunctionName;
		}
	}
	
	//==========================================================
	//
	//	Class		:	FixedRect
	//
	//	Summary		:	Fixed position and size
	//
	//==========================================================
	jaswin.FixedRect = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iAnchorX,		// * Horizontal anchor[int]
					iAnchorY,		// * Vertical anchor[int]
					iPosX,			// * Horizontal position[int]
					iPosY,			// * Vertical position[int]
					iWidth,			// * Width[int]
					iHeight)		// * Height[int]
		{
			// Check parameters
			if((iAnchorX != jaswin.ANCHOR_LEFT)
			&& (iAnchorX != jaswin.ANCHOR_RIGHT))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ANCHOR_X,
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			if((iAnchorY != jaswin.ANCHOR_TOP)
			&& (iAnchorY != jaswin.ANCHOR_BOTTOM))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ANCHOR_Y,
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			if(typeof iPosX != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_X, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			if(typeof iPosY != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_Y, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			if(typeof iWidth != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			if(iWidth < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_WIDTH, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			if(typeof iHeight != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			if(iHeight < 0)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_HEIGHT, ""),
											"jaswin.js", "jaswin.FixedRect.constructor");
			
			// Set properties
			this.anchorX = iAnchorX;
			this.anchorY = iAnchorY;
			this.posX = iPosX;
			this.posY = iPosY;
			this.width = iWidth;
			this.height = iHeight;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	ExpandableRect
	//
	//	Summary		:	Expandable position and size
	//
	//==========================================================
	jaswin.ExpandableRect = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iAnchorX,		// * Horizontal anchor[int]
					iAnchorY,		// * Vertical anchor[int]
					iLeft,			// Left position(ANCHOR_BOTH or ANCHOR_LEFT)
					iRight,			// Right position(ANCHOR_BOTH or ANCHOR_RIGHT)
					iTop,			// Top position(ANCHOR_BOTH or ANCHOR_TOP)
					iBottom,		// Bottom position(ANCHOR_BOTH or ANCHOR_BOTTOM)
					iWidth,			// Width(except ANCHOR_BOTH)
					iHeight)		// Height(except ANCHOR_BOTH)
		{
			// Check parameters
			if((iAnchorX != jaswin.ANCHOR_LEFT)
			&& (iAnchorX != jaswin.ANCHOR_RIGHT)
			&& (iAnchorX != jaswin.ANCHOR_BOTH))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ANCHOR_X,
											"jaswin.js", "jaswin.ExpandableRect.constructor");
			
			if((iAnchorY != jaswin.ANCHOR_TOP)
			&& (iAnchorY != jaswin.ANCHOR_BOTTOM)
			&& (iAnchorY != jaswin.ANCHOR_BOTH))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ANCHOR_Y,
											"jaswin.js", "jaswin.ExpandableRect.constructor");
			
			if((iAnchorX == jaswin.ANCHOR_LEFT) || (iAnchorX == jaswin.ANCHOR_BOTH))
			{
				if(typeof iLeft != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_LEFT,
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			if((iAnchorX == jaswin.ANCHOR_RIGHT) || (iAnchorX == jaswin.ANCHOR_BOTH))
			{
				if(typeof iRight != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_RIGHT,
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			if((iAnchorY == jaswin.ANCHOR_TOP) || (iAnchorY == jaswin.ANCHOR_BOTH))
			{
				if(typeof iTop != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_TOP,
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			if((iAnchorY == jaswin.ANCHOR_BOTTOM) || (iAnchorY == jaswin.ANCHOR_BOTH))
			{
				if(typeof iBottom != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_BOTTOM,
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			if(iAnchorX != jaswin.ANCHOR_BOTH)
			{
				if(typeof iWidth != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_WIDTH, ""),
												"jaswin.js", "jaswin.ExpandableRect.constructor");
				if(iWidth <= 0)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_WIDTH, ""),
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			if(iAnchorY != jaswin.ANCHOR_BOTH)
			{
				if(typeof iHeight != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_HEIGHT, ""),
												"jaswin.js", "jaswin.ExpandableRect.constructor");
				if(iHeight <= 0)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_HEIGHT, ""),
												"jaswin.js", "jaswin.ExpandableRect.constructor");
			}
			
			// Set properties
			this.anchorX = iAnchorX;
			this.anchorY = iAnchorY;
			this.left = iLeft;
			this.right = iRight;
			this.top = iTop;
			this.bottom = iBottom;
			this.width = iWidth;
			this.height = iHeight;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	Control
	//
	//	Summary		:	Base class of normal type controls
	//
	//==========================================================
	jaswin.Control = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(bAcceptNoID,	// * Accept no ID or not[boolean]
					sID,			// * ID[string]
					bContainer,		// * Container or not[boolean]
					oRect,			// * Size and position[FixedRect / ExpandableRect]
					bEnabled,		// * Enabled[boolean]
					iTabIndex)		// Tab index[int]
		{
			// Check parameters
			if(bAcceptNoID)
			{
				if(sID)
				{
					if(typeof sID != "string")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ID,
													"jaswin.js", "jaswin.Control.constructor");
				}
			}
			else
			{
				if(typeof sID != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_ID,
												"jaswin.js", "jaswin.Control.constructor");
			}
			
			if(typeof bContainer != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTAINER, sID),
											"jaswin.js", "jaswin.Control.constructor");
			
			if((typeof oRect != "object")
			|| (!bContainer && (!oRect instanceof jaswin.FixedRect))
			|| (bContainer && (!oRect instanceof jaswin.ExpandableRect)))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_RECT, sID),
											"jaswin.js", "jaswin.Control.constructor");
			
			if(typeof bEnabled != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ENABLED, sID),
											"jaswin.js", "jaswin.Control.constructor");
			
			if(iTabIndex)
			{
				if(typeof iTabIndex != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TABINDEX, sID),
												"jaswin.js", "jaswin.Control.constructor");
			}
			
			// Set properties
			this.id = sID;					// ID[String]
			if(!this.id)
				this.id = "";
			this.rect = oRect;				// Size and position[FixedRect / ExpandableRect]
			this.element = null;			// Element[DOM Element]
			this.focusElement = null;		// Focus element[DOM Element]
			this.parent = null;				// Parent class[global / window / panel class]
			this.enabled = bEnabled;		// Enabled or not[boolean]
			this.parentEnabled = true;		// Enabled or not[boolean]
			if(iTabIndex)
				this.tabIndex = iTabIndex;
			else
				this.tabIndex = -1;
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			let oParentWnd = jaswin.findParentWindow(this);
			
			this.element.parentNode.removeChild(this.element);
			delete oParentWnd[this.id];
		}
		
		//==================================================
		//	Set this control to parent class
		//==================================================
		setParent(oParent)
		{
			if((typeof oParent != "object")
			|| (!oParent instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_PARENT, this.id),
											"jaswin.js", "jaswin.Control.setParent");
			
			this.parent = oParent;
			if((this.parent.enabled) && (!this.parentEnabled))
				this.enable(true);
			else if((!this.parent.enabled) && (this.parentEnabled))
				this.disable(true);
			
			if((this.id) && (this.id.length > 0))
			{
				let oParentWnd = jaswin.findParentWindow(this);
				let sID = "";
				
				if(oParentWnd == global)
					sID = this.id;
				else
					sID = oParentWnd.id + "_" + this.id;
				
				// Duplication check
				if(document.getElementById(sID))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_DUPLICATION_ID, this.id),
												"jaswin.js", "jaswin.Control.setParent");
				
				this.element.id = sID;
				oParentWnd[this.id] = this;
			}
		}
		
		//==================================================
		//	Enable this control
		//==================================================
		enable(bParent)
		{
			if(bParent)
				this.parentEnabled = true;
			else
				this.enabled = true;
			
			if((this.enabled) && (this.focusElement))
				this.focusElement.tabIndex = this.tabIndex;
		}
		
		//==================================================
		//	Disable this control
		//==================================================
		disable(bParent)
		{
			if(bParent)
				this.parentEnabled = false;
			else
				this.enabled = false;
			
			if(this.focusElement)
				this.focusElement.tabIndex = -1;
		}
		
		//==================================================
		//	Show this control
		//==================================================
		show()
		{
			if(this == global)
				return;
			this.element.style.display = "block";
		}
		
		//==================================================
		//	Hide this control
		//==================================================
		hide()
		{
			if(this == global)
				return;
			this.element.style.display = "none";
		}
		
		//==================================================
		//	Visible or not
		//==================================================
		isVisible()
		{
			if(this.element.style.display == "none")
				return false;
			else
				return true;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	Container
	//
	//	Summary		:	Base class of Container type controls
	//
	//==========================================================
	jaswin.Container = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,		// * ID[string]
					oRect,		// * Size and position[ExpandableRect]
					bEnabled,	// * Enabled[boolean]
					iTabIndex)	// Tab index[int]
		{
			// Base class
			super(false, sID, true, oRect, bEnabled, iTabIndex);
			
			// Set properties
			this.container = null;			// Container for client area[DOM Element]
			this.controls = new Array();	// Controls on this panel[Array of Control classes]
			this.contextMenu = null;		// Context menu[Menu class]
			this.onContextMenu = null;		// Function on right click[function]
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
				this.controls[iCounter].destroy();
			
			if(this.contextMenu)
				this.contextMenu.destroy();
			
			super.destroy();
		}
		
		//==================================================
		//	Enable this container
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled && this.parentEnabled)
			{
				let iCounter;
				
				for(iCounter = 0; iCounter < this.controls.length; iCounter++)
					this.controls[iCounter].enable(true);
			}
		}
		
		//==================================================
		//	Disable this container
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			let iCounter;
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.controls[iCounter] instanceof jaswin.BaseWindow)
					continue;
				else
					this.controls[iCounter].disable(true);
			}
		}
		
		//==================================================
		//	Add control to this container
		//==================================================
		addControl(oControl, oParentElement)
		{
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, this.id),
											"jaswin.js", "jaswin.Container.addControl");
			
			if(oParentElement)
			{
				if(!oParentElement.getElementsByTagName)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_PARENT_ELEMENT, this.id),
												"jaswin.js", "jaswin.Container.addControl");
				
				if(!this.element.contains(oParentElement))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_PARENT_ELEMENT, this.id),
												"jaswin.js", "jaswin.Container.addControl");
			}
			
			if(!this.parent)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_CONTROL_ADD, this.id),
											"jaswin.js", "jaswin.Container.addControl");
			
			let oParentWnd = jaswin.findParentWindow(this);
			if((jaswin.frame) && (oControl instanceof jaswin.frame.Window))
			{
				if(oParentWnd != global)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_WINDOW_ALLOCATION, this.id),
												"jaswin.js", "jaswin.Container.addControl");
			}
			
			// Append element and set parent
			if(oParentElement)
				oParentElement.appendChild(oControl.element);
			else
				this.container.appendChild(oControl.element);
			this.controls.push(oControl);
			oControl.setParent(this);
			
			if((jaswin.frame)
			&& (oControl instanceof jaswin.frame.Window)
			&& (oControl.isVisible()))
			{
				// Activate window
				oControl.setActive();
			}
			
			// Set tabstop
			if((oControl.tabIndex) && (oControl.tabIndex > 0))
			{
				oParentWnd.addTabstop(oControl);
				if(oParentWnd != jaswin.currentWindow)
					oControl.focusElement.tabIndex = -1;
			}
		}
		
		//==================================================
		//	Remove control from this container
		//==================================================
		removeControl(oControl)
		{
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, this.id),
											"jaswin.js", "jaswin.Container.removeControl");
			
			// Remove tabstop
			if((oControl.tabIndex) && (oControl.tabIndex > 0))
			{
				let oParentWnd = jaswin.findParentWindow(this);
				oParentWnd.removeTabstop(oControl);
			}
			
			// Remove control
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.controls[iCounter] == oControl)
				{
					this.controls.splice(iCounter, 1);
					oControl.destroy();
					return;
				}
			}
		}
		
		//==================================================
		//	Set context menu to this container
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			if(this != global)
			{
				if(!this.parent)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_CONTROL_ADD, this.id),
												"jaswin.js", "jaswin.Container.setContextMenu");
			}
			if(!jaswin.menu)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "menu.js"),
											"jaswin.js", "jaswin.Container.setContextMenu");
			
			if((typeof oMenu != "object")
			|| (!oMenu instanceof jaswin.menu.Menu))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MENU, this.id),
											"jaswin.js", "jaswin.Container.setContextMenu");
			
			if(fpContext)
			{
				if(typeof fpContext != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CONTEXTMENU, this.id),
												"jaswin.js", "jaswin.Container.setContextMenu");
			}
			
			if(this.contextMenu)
			{
				this.contextMenu.destroy();
				this.contextMenu = null;
			}
			
			this.contextMenu = oMenu;
			
			if(fpContext)
				this.onContextMenu = fpContext;
		}
		
		//==================================================
		//	Remove context menu from this container
		//==================================================
		removeContextMenu()
		{
			if(this.contextMenu)
			{
				this.contextMenu.destroy();
				this.contextMenu = null;
			}
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	BaseWindow
	//
	//	Summary		:	Base class of Window type controls
	//
	//==========================================================
	jaswin.BaseWindow = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,		// * ID[string]
					oRect)		// * Size and position[ExpandableRect]
		{
			super(sID, oRect, true);
			
			this.tabstops = new Array();		// Controls for tabstop[Array of Control classes]
			this.menu = null;					// Main menu on this window[MenuBar class]
			this.toolbar = null;				// Toolbar on global[ToolBar class]
			this.active = true;					// Active window or not[boolean]
			this.onResize = null;				// Function on resize[function]
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			if(this == global)
				return;
			
			if(this.menu)
				this.menu.destroy();
			
			if(this.toolbar)
				this.toolbar.destroy();
			
			super.destroy();
		}
		
		//==================================================
		//	Enable this window
		//==================================================
		enable(bParent)
		{
			return;
		}
		
		//==================================================
		//	Disable this window
		//==================================================
		disable(bParent)
		{
			return;
		}
		
		//==================================================
		//	Set active
		//==================================================
		setActive()
		{
			this.active = true;
			jaswin.currentWindow = this;
			
			let iCounter = 0;
			let bFirst = true;
			
			for(iCounter = 0; iCounter < this.tabstops.length; iCounter++)
			{
				if((this.tabstops[iCounter].enabled) && (this.tabstops[iCounter].parentEnabled))
				{
					this.tabstops[iCounter].focusElement.tabIndex = this.tabstops[iCounter].tabIndex;
					if(bFirst)
					{
						this.tabstops[iCounter].focusElement.focus();
						bFirst = false;
					}
				}
			}
			
			if(this == global)
			{
				if(jaswin.frame)
				{
					for(let sKey in jaswin.frame.windows)
						jaswin.frame.windows[sKey].setInactive();
				}
			}
			else
			{
				global.setInactive();
				if(jaswin.frame)
				{
					let iCurIndex = 0;
					let iTmpIndex = 0;
					
					iCurIndex = this.zIndex;
					for(let sKey in jaswin.frame.windows)
					{
						if(jaswin.frame.windows[sKey] == this)
						{
							this.zIndex = jaswin.frame.windowCount;
							this.element.style.zIndex = jaswin.frame.windowCount + this.modalLevel;
						}
						else
						{
							iTmpIndex = jaswin.frame.windows[sKey].zIndex;
							if(jaswin.frame.windows[sKey].zIndex > iCurIndex)
							{
								iTmpIndex--;
								jaswin.frame.windows[sKey].zIndex = iTmpIndex;
								jaswin.frame.windows[sKey].element.style.zIndex = iTmpIndex + jaswin.frame.windows[sKey].modalLevel;
							}
							jaswin.frame.windows[sKey].setInactive();
						}
					}
				}
			}
		}
		
		//==================================================
		//	Set inctive
		//==================================================
		setInactive()
		{
			this.active = false;
			
			let iCounter = 0;
			
			for(iCounter = 0; iCounter < this.tabstops.length; iCounter++)
				this.tabstops[iCounter].focusElement.tabIndex = -1;
			
			if(this.menu)
				this.menu.setInactive();
		}
		
		//==================================================
		//	Set this control to parent class
		//==================================================
		setParent(oParent)
		{
			if(this == global)
				return;
		}
		
		//==================================================
		//	Add control to this window
		//==================================================
		addControl(oControl, oParentElement)
		{
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, this.id),
											"jaswin.js", "jaswin.BaseWindow.addControl");
			
			if(oParentElement)
			{
				if(!oParentElement.getElementsByTagName)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_PARENT_ELEMENT, this.id),
												"jaswin.js", "jaswin.BaseWindow.addControl");
				
				if(!this.element.contains(oParentElement))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_PARENT_ELEMENT, this.id),
												"jaswin.js", "jaswin.BaseWindow.addControl");
			}
			
			if(this != global)
			{
				if(!this.parent)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_CONTROL_ADD, this.id),
												"jaswin.js", "jaswin.BaseWindow.addControl");
				
				if((jaswin.frame) && (oControl instanceof jaswin.frame.Window))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_WINDOW_ALLOCATION, this.id),
												"jaswin.js", "jaswin.BaseWindow.addControl");
			}
			
			// Append element and set parent
			if(oParentElement)
				oParentElement.appendChild(oControl.element);
			else
				this.container.appendChild(oControl.element);
			this.controls.push(oControl);
			oControl.setParent(this);
			
			if((jaswin.frame)
			&& (oControl instanceof jaswin.frame.Window)
			&& (oControl.isVisible()))
			{
				// Activate window
				oControl.setActive();
			}
			
			// Set tabstop
			if((oControl.tabIndex) && (oControl.tabIndex > 0))
			{
				this.addTabstop(oControl);
				if(jaswin.currentWindow != this)
					oControl.focusElement.tabIndex = -1;
			}
		}
		
		//==================================================
		//	Remove control from this window
		//==================================================
		removeControl(oControl)
		{
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, this.id),
											"jaswin.js", "jaswin.BaseWindow.removeControl");
			
			// Remove tabstop
			if((oControl.tabIndex) && (oControl.tabIndex > 0))
				this.removeTabstop(oControl);
			
			// Remove control
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.controls[iCounter] == oControl)
				{
					this.controls.splice(iCounter, 1);
					oControl.destroy();
					return;
				}
			}
		}
		
		//==================================================
		//	Add tabstop to window
		//==================================================
		addTabstop(oControl)
		{
			if((typeof oControl != "object")
			|| (!oControl instanceof jaswin.Control))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, this.id),
											"jaswin.js", "jaswin.Window.addTabstop");
			
			// Set tabstop
			if((oControl.tabIndex) && (oControl.tabIndex > 0))
			{
				this.tabstops.push(oControl);
				if(this.tabstops.length > 1)
					this.tabstops.sort(jaswin.compareTabstop);
			}
		}
		
		//==================================================
		//	Remove tabstop from window
		//==================================================
		removeTabstop(oControl)
		{
			let iCounter;
			
			for(iCounter = 0; iCounter < this.tabstops.length; iCounter++)
			{
				if(this.tabstops[iCounter] == oControl)
				{
					this.tabstops.splice(iCounter, 1);
					return;
				}
			}
		}
		
		//==================================================
		//	Focus previous tabstop element
		//==================================================
		prevTabstop(e)
		{
			let iCounter;
			let iIndex = -1;
			let iMinIndex = -1;
			let oaTarget = new Array();
			
			for(iCounter = 0; iCounter < this.tabstops.length; iCounter++)
			{
				if((this.tabstops[iCounter].element.offsetWidth != 0)
				&& (this.tabstops[iCounter].enabled && this.tabstops[iCounter].parentEnabled))
				{
					oaTarget.push(this.tabstops[iCounter]);
					if(this.tabstops[iCounter].element.contains(document.activeElement))
						iIndex = oaTarget.length - 1;
					if(iMinIndex == -1)
						iMinIndex = oaTarget.length - 1;
				}
			}
			if(oaTarget.length <= 1)
			{
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			
			if((iIndex == -1) || (iIndex == iMinIndex))
				iIndex = oaTarget.length - 1;
			else
				iIndex--;
			
			oaTarget[iIndex].focusElement.focus();
			e.preventDefault();
			e.stopPropagation();
			return false;
		}
		
		//==================================================
		//	Focus next tabstop element
		//==================================================
		nextTabstop(e)
		{
			let iCounter;
			let iIndex = -1;
			let iMaxIndex = -1;
			let oaTarget = new Array();
			
			for(iCounter = 0; iCounter < this.tabstops.length; iCounter++)
			{
				if((this.tabstops[iCounter].element.offsetWidth != 0)
				&& (this.tabstops[iCounter].enabled && this.tabstops[iCounter].parentEnabled))
				{
					oaTarget.push(this.tabstops[iCounter]);
					if(this.tabstops[iCounter].element.contains(document.activeElement))
						iIndex = oaTarget.length - 1;
					if(iMaxIndex < oaTarget.length - 1)
						iMaxIndex = oaTarget.length - 1;
				}
			}
			if(oaTarget.length <= 1)
			{
				e.preventDefault();
				e.stopPropagation();
				return false;
			}
			
			if((iIndex == -1) || (iIndex == iMaxIndex))
				iIndex = 0;
			else
				iIndex++;
			
			oaTarget[iIndex].focusElement.focus();
			e.preventDefault();
			e.stopPropagation();
			return false;
		}
		
		//==================================================
		//	Set main menu
		//==================================================
		setMenu(oMenuBar)
		{
			if(!jaswin.menu)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "menu.js"),
											"jaswin.js", "jaswin.Window.setMenu");
			
			if((typeof oMenuBar != "object")
			|| (!oMenuBar instanceof jaswin.menu.MenuBar))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MENU_HEADER, this.id),
											"jaswin.js", "jaswin.Window.setMenu");
			
			this.menu = oMenuBar;
		}
		
		//==================================================
		//	Set toolbar
		//==================================================
		setToolBar(oToolBar)
		{
			if(!jaswin.toolbar)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "toolbar.js"),
											"jaswin.js", "jaswin.Window.setToolBar");
			
			if((typeof oToolBar != "object")
			|| (!oToolBar instanceof jaswin.toolbar.ToolBar))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLBAR, this.id),
											"jaswin.js", "jaswin.Window.setToolBar");
			
			this.toolbar = oToolBar;
		}
		
		//==================================================
		//	Set context menu to window
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			this.container.appendChild(oMenu.element);
			this.contextMenu = oMenu;
			oMenu.setParent(this, 0);
		}
		
		//==================================================
		//	Show context menu
		//==================================================
		showContextMenu(e)
		{
			let oElement = e.target;
			
			if(oElement != this.container)
				return;
			
			if(this.onContextMenu)
			{
				if(!this.onContextMenu(this))
					return;
			}
			
			this.contextMenu.showContext(e);
		}
		
		//==================================================
		//	Set resize function
		//==================================================
		setResizeEvent(fpResize)
		{
			if(typeof fpResize != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_RESIZE, this.id),
											"jaswin.js", "jaswin.BaseWindow.setResizeEvent");
			this.onResize = fpResize;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	Global
	//
	//	Summary		:	Global class
	//
	//==========================================================
	jaswin.Global = class extends jaswin.BaseWindow
	{
		//==================================================
		// constructor
		//==================================================
		constructor()
		{
			let oRect = new jaswin.ExpandableRect(jaswin.ANCHOR_BOTH,
													jaswin.ANCHOR_BOTH,
													0, 0, 0, 0, 0, 0);
			super("global", oRect, true);
			
			this.element = document.createElement("div");		// Base element[DOM Element]
			this.element.className = "base";
			document.body.appendChild(this.element);
			this.container = document.createElement("div");		// Container element[DOM Element]
			this.container.id = this.id;
			this.container.className = "global";
			this.element.appendChild(this.container);
			
			// Event binding
			this.element.addEventListener("mousedown", (e) => {
				global.setActive();
				e.stopPropagation();
			}, false);
			window.addEventListener("resize", (e) => {
				global.resize();
			}, false);
			
			jaswin.currentWidth = this.container.clientWidth;
			jaswin.currentHeight = this.container.clientHeight;
			jaswin.lastWidth = this.container.clientWidth;
			jaswin.lastHeight = this.container.clientHeight;
			jaswin.currentWindow = this;
			
			document.body.addEventListener("keydown", keyControl, false);
		}
		
		//==================================================
		//	Set active
		//==================================================
		setActive()
		{
			super.setActive();
			if(!this.globalKeyEvent)
			{
				this.element.addEventListener("keydown", this.globalKeyEvent = (e) => {
					if(jaswin.currentWindow != global)
						return true;
					
					if(e.keyCode == 9)
					{
						// Tab
						if(e.shiftKey)
							return global.prevTabstop(e);
						else
							return global.nextTabstop(e);
					}
					return true;
				}, false);
			}
		}
		
		//==================================================
		//	Set inctive
		//==================================================
		setInactive()
		{
			super.setInactive();
			if(this.globalKeyEvent)
			{
				this.element.removeEventListener("keydown", this.globalKeyEvent, false);
				this.globalKeyEvent = null;
			}
		}
		
		//==================================================
		//	Set main menu
		//==================================================
		setMenu(oMenuBar)
		{
			super.setMenu(oMenuBar);
			
			this.element.appendChild(this.menu.element);
			this.menu.setParent(this)
			
			let iTop = 22;
			
			if(this.toolbar)
				iTop += parseInt(this.toolbar.style.height) + 2;
			
			this.container.style.top = iTop + "px";
			
			this.element.addEventListener("menu", (e) => {
				let iKeyCode = 0;
				let sChar = "";
				
				if(!this.menu) return;
				
				iKeyCode = e.detail;
				sChar = String.fromCharCode(iKeyCode);
				this.menu.shortCut(sChar);
			}, false);
			
			this.resize();
		}
		
		//==================================================
		//	Set toolbar
		//==================================================
		setToolBar(oToolBar)
		{
			super.setToolBar(oToolBar);
			
			let iTop = 0;
			let iOffset = 0;
			
			if(this.menu)
				iTop += 22;
			
			this.toolbar.element.style.top = iTop + "px";
			this.element.appendChild(this.toolbar.element);
			this.toolbar.setParent(this);
			
			if(this.toolbar.controls.length > 0)
				iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
			
			this.toolbar.element.style.height = 20 + iOffset + "px";
			iTop += 22 + iOffset;
			
			this.container.style.top = iTop + "px";
			
			this.resize();
		}
		
		//==================================================
		//	Show tool bar
		//==================================================
		showToolBar()
		{
			if(!this.toolbar)
				return;
			this.toolbar.show();
			
			let iTop = 0;
			let iOffset = 0;
			
			if(this.menu)
				iTop += 22;
			
			if(this.toolbar.controls.length > 0)
				iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
			
			this.toolbar.element.style.height = 20 + iOffset + "px";
			iTop += 22 + iOffset;
			
			this.container.style.top = iTop + "px";
			this.resize();
		}
		
		//==================================================
		//	Hide tool bar
		//==================================================
		hideToolBar()
		{
			if(!this.toolbar)
				return;
			this.toolbar.hide();
			
			let iTop = 0;
			
			if(this.menu)
				iTop += 22;
			
			this.container.style.top = iTop + "px";
			this.resize();
		}
		
		//==================================================
		//	Notify destroy window
		//==================================================
		notifyDestroyWindow(oWindow)
		{
			if((typeof oWindow != "object")
			|| (!oWindow instanceof jaswin.frame.Window))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WINDOW, this.id),
											"jaswin.js", "jaswin.Global.notifyDestroyWindow");
			
			let iCounter;
			
			for(iCounter = 0; iCounter < this.controls.length; iCounter++)
			{
				if(this.controls[iCounter] == oWindow)
				{
					this.controls.splice(iCounter, 1);
					return;
				}
			}
		}
		
		//==================================================
		//	Set context menu to global
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			if(!this.globalContextMenu)
			{
				this.container.addEventListener("contextmenu", this.globalContextMenu = (e) => {
					this.showContextMenu(e);
					return false;
				}, false);
			}
		}
		
		//==================================================
		//	Remove context menu from global
		//==================================================
		removeContextMenu()
		{
			super.removeContextMenu();
			if(this.globalContextMenu)
				this.container.removeEventListener("contextmenu", this.globalContextMenu, false);
			this.globalContextMenu = null;
		}
		
		//==================================================
		//	Resize event
		//==================================================
		resize()
		{
			if((this.toolbar) && this.toolbar.isVisible())
			{
				let iTop = 0;
				let iOffset = 0;
				
				if(this.menu)
					iTop += 22;
				
				if(this.toolbar.controls.length > 0)
					iOffset = this.toolbar.controls[this.toolbar.controls.length - 1].element.offsetTop;
				
				this.toolbar.element.style.height = 20 + iOffset + "px";
				iTop += 22 + iOffset;
				
				this.container.style.top = iTop + "px";
			}
			
			jaswin.lastWidth = jaswin.currentWidth;
			jaswin.lastHeight = jaswin.currentHeight;
			jaswin.currentWidth = this.container.clientWidth;
			jaswin.currentHeight = this.container.clientHeight;
			
			if(this.onResize)
				this.onResize();
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	keyControl
	//
	//	Summary		:	Keyboard  control of html body
	//
	//==========================================================
	function keyControl(e)
	{
		let oTarget = null;
		let oEvent = null;
		
		if(e.altKey)
		{
			e.preventDefault();
			
			oTarget = jaswin.currentWindow;
			
			if(!oTarget.menu)
			{
				if((oTarget != global) && (global.menu))
					oTarget = global;
				else
					return false;
			}
			
			oEvent = document.createEvent("UIEvents");
			oEvent.initUIEvent("menu", false, true, document.defaultView, e.keyCode);
			
			oTarget.element.dispatchEvent(oEvent);
			
			return false;
		}
		else
		{
			if(jaswin.menu)
			{
				let iCnt = jaswin.menu.visibleMenus.length;
				
				if(iCnt > 0)
				{
					// A menu is visible
					e.preventDefault();
					
					oEvent = document.createEvent("UIEvents");
					oEvent.initUIEvent("menuKey", false, true, document.defaultView, e.keyCode);
					jaswin.menu.visibleMenus[iCnt - 1].element.dispatchEvent(oEvent);
					
					return false;
				}
			}
		}
		return true;
	}
	
	
	//==========================================================
	//
	//	Function	:	getBrowserName
	//
	//	Summary		:	Get name of this browser
	//
	//==========================================================
	jaswin.getBrowserName = function()
	{
		let sUserAgent = window.navigator.userAgent.toLowerCase();
		
		if((sUserAgent.indexOf("opera") != -1) || (sUserAgent.indexOf("opr") != -1))
			jaswin.BROWSER = "opera";
		else if (sUserAgent.indexOf("edge") != -1)
			jaswin.BROWSER = "edge";
		else if (sUserAgent.indexOf("chrome") != -1)
			jaswin.BROWSER = "chrome";
		else if (sUserAgent.indexOf("msie") != -1)
			jaswin.BROWSER = "";
		else if (sUserAgent.indexOf("safari") != -1)
			jaswin.BROWSER = "safari";
		else if (sUserAgent.indexOf("like gecko") != -1)
			jaswin.BROWSER = "";
		else if (sUserAgent.indexOf("gecko") != -1)
			jaswin.BROWSER = "gecko";
		else
			jaswin.BROWSER = "";
	}
	
	
	//==========================================================
	//
	//	Function	:	formatMessage
	//
	//	Summary		:	Format a message
	//
	//==========================================================
	jaswin.formatMessage = function()
	{
		if((!arguments) || (arguments.length == 0))
			return "";
		
		let sRet = arguments[0];
		let iCounter;
		
		if(!sRet)
			return "";
		
		for(iCounter = 1; iCounter < arguments.length; iCounter++)
			sRet = sRet.replace(new RegExp("\\{" + iCounter + "\\}", "g"), arguments[iCounter]);
		
		sRet = sRet.replace(new RegExp("\\{\\d+\\}", "g"), "");
		
		return sRet;
	}
	
	
	//==========================================================
	//
	//	Function	:	compareTabstop
	//
	//	Summary		:	Compare tab stop for sort
	//
	//==========================================================
	jaswin.compareTabstop = function(oControl1, oControl2)
	{
		if(!oControl2.tabIndex)
			return -1;
		if(!oControl1.tabIndex)
			return 1;
		
		if(oControl2.tabIndex < 0)
			return -1;
		if(oControl1.tabIndex < 0)
			return 1;
		
		if(oControl1.tabIndex < oControl2.tabIndex)
			return -1;
		else if(oControl1.tabIndex > oControl2.tabIndex)
			return 1;
		
		return 0;
	}
	
	
	//==========================================================
	//
	//	Function	:	findParentWindow
	//
	//	Summary		:	Find parent window
	//
	//==========================================================
	jaswin.findParentWindow = function(oControl)
	{
		if((typeof oControl != "object")
		|| ((!oControl instanceof jaswin.Control) && (!oControl instanceof jaswin.Container)))
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_TYPE_CONTROL, ""),
										"jaswin.js", "jaswin.findParentWindow");
		let oCurClass = oControl;
		
		if(!jaswin.frame)
			return global;
		while(true)
		{
			if(!oCurClass.parent)
				return global;
			
			if(oCurClass.parent == global)
				return global;
			
			if(oCurClass.parent instanceof jaswin.frame.Window)
				return oCurClass.parent;
			
			oCurClass = oCurClass.parent;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	createXML
	//
	//	Summary		:	Create XML DOM object
	//
	//==========================================================
	jaswin.createXML = function(sDocument)	// * Document text[String]
	{
		if(typeof sDocument != "string")
			throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_DOC_TEXT,
										"jaswin.js", "jaswin.createXML");
		
		let oParser = new DOMParser();
		let oXml = null;
		
		try
		{
			oXml = oParser.parseFromString(sDocument, "application/xml");
		}
		catch(e)
		{
			throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_XML_PARSER,
										"jaswin.js", "jaswin.createXML");
		}
		
		return oXml.documentElement;
	}
	
	
	//==========================================================
	//
	//	Function	:	escapeText
	//
	//	Summary		:	Escape HTML/XML tag characters
	//
	//==========================================================
	jaswin.escapeText = function(sOrg)		// Original text[string]
	{
		let sRet = sOrg.replace(/&/g, "&amp;");
		sRet = sRet.replace(/"/g, "&quot;");
		sRet = sRet.replace(/'/g, "&#039;");
		sRet = sRet.replace(/</g, "&lt;");
		sRet = sRet.replace(/>/g, "&gt;");
		
		return sRet;
	}
	
	
	//==========================================================
	//
	//	Function	:	unescapeText
	//
	//	Summary		:	Unescape HTML/XML tag characters
	//
	//==========================================================
	jaswin.unescapeText = function(sOrg)	// Original text[string]
	{
		let sRet = sOrg.replace(/&amp;/g, "&");
		sRet = sRet.replace(/&quot;/g, "\"");
		sRet = sRet.replace(/&#039;/g, "'");
		sRet = sRet.replace(/&lt;/g, "<");
		sRet = sRet.replace(/&gt;/g, ">");
		
		return sRet;
	}
	
	
	//==========================================================
	//
	//	Function	:	escapeRegExp
	//
	//	Summary		:	Escape letters for regular expression
	//
	//==========================================================
	jaswin.escapeRegExp = function(sOrg)	// Original text[string]
	{
		let sRet = "";
		let iCounter;
		
		for(iCounter = 0; iCounter < sOrg.length; iCounter++)
		{
			if(sOrg.charAt(iCounter) == "\\")
				sRet += "\\\\";
			else if(sOrg.charAt(iCounter) == "*")
				sRet += "\\*";
			else if(sOrg.charAt(iCounter) == "+")
				sRet += "\\+";
			else if(sOrg.charAt(iCounter) == ".")
				sRet += "\\.";
			else if(sOrg.charAt(iCounter) == "?")
				sRet += "\\?";
			else if(sOrg.charAt(iCounter) == "{")
				sRet += "\\{";
			else if(sOrg.charAt(iCounter) == "}")
				sRet += "\\}";
			else if(sOrg.charAt(iCounter) == "(")
				sRet += "\\(";
			else if(sOrg.charAt(iCounter) == ")")
				sRet += "\\)";
			else if(sOrg.charAt(iCounter) == "[")
				sRet += "\\[";
			else if(sOrg.charAt(iCounter) == "]")
				sRet += "\\]";
			else if(sOrg.charAt(iCounter) == "^")
				sRet += "\\^";
			else if(sOrg.charAt(iCounter) == "$")
				sRet += "\\$";
			else if(sOrg.charAt(iCounter) == "|")
				sRet += "\\|";
			else
				sRet += sOrg.charAt(iCounter);
		}
		return sRet;
	}
	
	
	//==========================================================
	//
	//	Function	:	formatNumber
	//
	//	Summary		:	Format numeric string
	//
	//==========================================================
	jaswin.formatNumber = function(sSrc,		// Source text[string]
									iDigit)		// Digit[int]
	{
		if(typeof sSrc != "string")
			throw new jaswin.JaswinError(jaswin.ERR_ALERT, ALERT_TYPE_SOURCE, "jaswin.js", "jaswin.formatNumber");
		if(typeof iDigit != "number")
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_TYPE_DIGIT, ""),
										"jaswin.js", "jaswin.formatNumber");
		if(iDigit < 0)
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_VALUE_DIGIT, ""),
										"jaswin.js", "jaswin.formatNumber");
		
		let sValue = sSrc;
		let iIndex = sValue.indexOf(".");
		let iCounter;
		let iUnder = 0;
		let iOver = 0;
		let iCnt = 0;
		let bMinus = false;
		let iStart = 0;
		let iEnd = 0;
		
		if(sValue.charAt(0) == "-")
		{
			bMinus = true;
			sValue = sValue.substring(1);
			iIndex = sValue.indexOf(".");
		}
		
		if((sValue.charAt(0) == "0") && (iIndex != 1))
		{
			if(iIndex == -1)
				iEnd = sValue.length;
			else
				iEnd = iIndex;
			
			for(iCounter = 0; iCounter < iEnd; iCounter++)
			{
				if(sValue.charAt(iCounter) != "0")
				{
					iStart = iCounter;
					break;
				}
			}
			sValue = sValue.substring(iStart);
			iIndex = sValue.indexOf(".");
		}
		
		// Fix digit
		if(iIndex == -1)
		{
			if(iDigit)
			{
				sValue += ".";
				for(iCounter = 0; iCounter < iDigit; iCounter++)
					sValue += "0";
			}
		}
		else
		{
			if(iDigit)
			{
				if(iIndex >= sValue.length - 1)
				{
					for(iCounter = 0; iCounter < iDigit; iCounter++)
						sValue += "0";
				}
				else
				{
					iUnder = sValue.substring(iIndex + 1).length;
					if(iUnder < iDigit)
					{
						for(iCounter = 0; iCounter < iDigit - iUnder; iCounter++)
							sValue += "0";
					}
					else if(iUnder > iDigit)
						sValue = sValue.substring(0, iIndex + iDigit + 1);
				}
				if(iIndex == 0)
					sValue = "0" + sValue;
			}
			else
			{
				if(iIndex == 0)
					sValue = "0";
				else
					sValue = sValue.substring(0, iIndex);
			}
		}
		
		// Comma
		iIndex = sValue.indexOf(".");
		if(iIndex == -1)
			iOver = sValue.length;
		else
			iOver = iIndex;
		if(iOver % 3)
			iCnt = Math.floor(iOver / 3);
		else
			iCnt = Math.floor(iOver / 3) - 1;
		
		for(iCounter = 0; iCounter < iCnt; iCounter++)
		{
			iIndex = sValue.lastIndexOf(",");
			if(iIndex == -1)
			{
				if(iOver % 3)
					sValue = sValue.substring(0, iOver % 3) + "," + sValue.substring(iOver % 3);
				else
					sValue = sValue.substring(0, 3) + "," + sValue.substring(3);
			}
			else
				sValue = sValue.substring(0, iIndex + 4) + "," + sValue.substring(iIndex + 4);
		}
		
		if(bMinus)
		{
			for(iCounter = 0; iCounter < sValue.length; iCounter++)
			{
				if((sValue.charAt(iCounter) != "0")
				&& (sValue.charAt(iCounter) != "."))
				{
					sValue = "-" + sValue;
					break;
				}
			}
		}
		
		return sValue;
	}
	
	
	//==========================================================
	//
	//	Function	:	parseFunction
	//
	//	Summary		:	Parse function
	//
	//==========================================================
	jaswin.parseFunction = function(fpFunc)		// Pointer of function
	{
		if(typeof fpFunc != "function")
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.ALERT_TYPE_FUNCTION,
										"jaswin.js", "jaswin.parseFunction");
		
		let sFuncSrc = fpFunc.toString();
		let saArgs = new Array();
		let iArgStart = sFuncSrc.indexOf("(");
		let iArgEnd = sFuncSrc.indexOf(")");
		let iStart = sFuncSrc.indexOf("{");
		
		if(iArgStart + 1 < iArgEnd)
			saArgs = sFuncSrc.substring(iArgStart + 1, iArgEnd).split(",");
		sFuncSrc = sFuncSrc.substring(iStart);
		
		return {name:fpFunc.name, args:saArgs, source:sFuncSrc};
	}
	
	
	//==========================================================
	//
	//	Function	:	messageBox
	//
	//	Summary		:	Show message box
	//
	//==========================================================
	jaswin.messageBox = function(oWindow,		// Return window on close[Window class]
								sMessage,		// Message text[string]
								sTitle,			// Title text[string]
								iStyle,			// * Style of message box[int]
								fpCallBack)		// Call back function[function]
	{
		// Check parameters
		if(!jaswin.frame)
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_INCLUDE, "frame.js"),
										"jaswin.js", "jaswin.messageBox");
		if(!jaswin.button)
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_INCLUDE, "button.js"),
										"jaswin.js", "jaswin.messageBox");
		
		let sMsg = "";
		let sCaption = "";
		
		if(oWindow)
		{
			if((typeof oWindow != "object")
			|| (!oWindow instanceof jaswin.frame.Window))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_WINDOW, "WndMessageBox"),
											"jaswin.js", "jaswin.messageBox");
		}
		
		if(sMessage)
		{
			if(typeof sMessage != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MESSAGE, "WndMessageBox"),
											"jaswin.js", "jaswin.messageBox");
			
			sMsg = jaswin.escapeText(sMessage);
			sMsg = sMsg.replace(/\r\n/g, "<br />");
			sMsg = sMsg.replace(/\r/g, "<br />");
			sMsg = sMsg.replace(/\n/g, "<br />");
		}
		
		if(sTitle)
		{
			if(typeof sTitle != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TITLE, "WndMessageBox"),
											"jaswin.js", "jaswin.messageBox");
			
			sCaption = sTitle;
		}
		
		if(typeof iStyle != "number")
			throw new jaswin.JaswinError(jaswin.ERR_ALERT,
										jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, "WndMessageBox"),
										"jaswin.js", "jaswin.messageBox");
		
		if(fpCallBack)
		{
			if(typeof fpCallBack != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_CALLBACK, "WND_MESSAGE_BOX"),
											"jaswin.js", "jaswin.messageBox");
		}
		
		// Create message box
		let oRect;
		let oBound;
		let oToolBarBound;
		let iWidth = 160;
		let iHeight = 100;
		let sID = "WndMessageBox";
		let iCnt = 0;
		let iAdjustTop = 0;
		
		if(oWindow)
		{
			oBound = oWindow.element.getBoundingClientRect();
			if(global.menu)
				iAdjustTop -= 22;
			if(global.toolbar)
			{
				oToolBarBound = global.toolbar.element.getBoundingClientRect();
				iAdjustTop -= oToolBarBound.height;
			}
			oRect = new jaswin.ExpandableRect(jaswin.ANCHOR_LEFT, jaswin.ANCHOR_TOP,
												oBound.left + Math.floor((oBound.width - iWidth) / 2), 0,
												oBound.top + Math.floor((oBound.height - iHeight) / 2) + iAdjustTop, 0,
												iWidth, iHeight);
		}
		else
		{
			oBound = global.container.getBoundingClientRect();
			oRect = new jaswin.ExpandableRect(jaswin.ANCHOR_LEFT, jaswin.ANCHOR_TOP,
												Math.floor((oBound.width - iWidth) / 2), 0,
												Math.floor((oBound.height - iHeight) / 2), 0,
												iWidth, iHeight);
		}
		
		while(true)
		{
			if((!document.getElementById(sID)) && (!jaswin.frame.windows[sID]))
				break;
			iCnt++;
			sID = "WndMessageBox" + iCnt;
		}
		
		let oMsgBox = new jaswin.frame.Window(sID, oRect, 0, 0,
												jaswin.frame.STYLE_TITLEBAR |
												jaswin.frame.STYLE_CLOSEBUTTON |
												jaswin.frame.STYLE_DESTROY_ON_CLOSE |
												jaswin.frame.STYLE_MOBABLE,
												sCaption,
												null);
		oMsgBox.returnValue = jaswin.MSGBOX_RETURN_CANCEL;
		
		// Icon
		let oSpan;
		let iMessageLeft = 16;
		
		if((iStyle & jaswin.MSGBOX_INFO)
		|| (iStyle & jaswin.MSGBOX_QUESTION)
		|| (iStyle & jaswin.MSGBOX_ALERT)
		|| (iStyle & jaswin.MSGBOX_ERROR))
		{
			oSpan = document.createElement("span");
			if(iStyle & jaswin.MSGBOX_INFO)
			{
				oSpan.className = jaswin.ICON_INFO;
				oSpan.innerHTML = "i";
			}
			else if(iStyle & jaswin.MSGBOX_QUESTION)
			{
				oSpan.className = jaswin.ICON_QUESTION;
				oSpan.innerHTML = "?";
			}
			else if(iStyle & jaswin.MSGBOX_ALERT)
			{
				oSpan.className = jaswin.ICON_ALERT;
				oSpan.innerHTML = "!";
			}
			else if(iStyle & jaswin.MSGBOX_ERROR)
			{
				oSpan.className = jaswin.ICON_ERROR;
				oSpan.innerHTML = "x";
			}
			oSpan.style.position = "absolute";
			oSpan.style.top = "8px";
			oSpan.style.left = "8px";
			
			oMsgBox.container.appendChild(oSpan);
			
			iMessageLeft = 48;
		}
		
		// Label
		oSpan = document.createElement("span");
		oSpan.className = "label";
		oSpan.style.left = iMessageLeft + "px";
		oSpan.style.top = "8px";
		oSpan.style.marginRight = "8px";
		oSpan.innerHTML = sMsg;
		oMsgBox.container.appendChild(oSpan);
		global.addControl(oMsgBox);
		
		// Button
		let oOkButton;
		let oCancelButton;
		let sButtonOk;
		let sButtonCancel;
		
		if(jaswin.MSGBOX_BUTTON_OK)
			sButtonOk = jaswin.MSGBOX_BUTTON_OK;
		else
			sButtonOk = "OK";
		if(jaswin.MSGBOX_BUTTON_CANCEL)
			sButtonCancel = jaswin.MSGBOX_BUTTON_CANCEL;
		else
			sButtonCancel = "Cancel";
		
		if((iStyle & jaswin.MSGBOX_OKCANCEL)
		|| (iStyle & jaswin.MSGBOX_YESNO))
		{
			if(iStyle & jaswin.MSGBOX_YESNO)
			{
				if(jaswin.MSGBOX_BUTTON_YES)
					sButtonOk = jaswin.MSGBOX_BUTTON_YES;
				else
					sButtonOk = "Yes";
				if(jaswin.MSGBOX_BUTTON_NO)
					sButtonCancel = jaswin.MSGBOX_BUTTON_NO;
				else
					sButtonCancel = "No";
			}
			
			oRect = new jaswin.FixedRect(jaswin.ANCHOR_RIGHT, jaswin.ANCHOR_BOTTOM,
										88, 8, 64, 20);
			oOkButton = new jaswin.button.Button("Btn" + sID + "_Ok", oRect,
												jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_TEXT,
												true,
												sButtonOk, null,
												1, messageBoxOk);
			oMsgBox.addControl(oOkButton);
			
			oRect = new jaswin.FixedRect(jaswin.ANCHOR_RIGHT, jaswin.ANCHOR_BOTTOM,
										8, 8, 72, 20);
			oCancelButton = new jaswin.button.Button("Btn" + sID + "_Cancel", oRect,
													jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_TEXT,
													true,
													sButtonCancel, null,
													2, messageBoxCancel);
			oMsgBox.addControl(oCancelButton);
		}
		else
		{
			oRect = new jaswin.FixedRect(jaswin.ANCHOR_RIGHT, jaswin.ANCHOR_BOTTOM,
										8, 8, 64, 20);
			oOkButton = new jaswin.button.Button("Btn" + sID + "_Ok", oRect,
												jaswin.button.STYLE_DEFAULT | jaswin.button.STYLE_TEXT,
												true,
												sButtonOk, null,
												1, messageBoxOk);
			oMsgBox.addControl(oOkButton);
		}
		oMsgBox.showDialog(oWindow, fpCallBack);
		oOkButton.element.focus();
		
		// Resize
		let iCounter;
		let iCurIndex = 0;
		let iLines = 1;
		let iGap = 0;
		
		if(oSpan.offsetWidth > oMsgBox.rect.width - iMessageLeft - 10)
		{
			iGap = oSpan.offsetWidth - (oMsgBox.rect.width - iMessageLeft - 10);
			if(iGap > 200)
			{
				oMsgBox.setSize(oMsgBox.rect.width + 200, oMsgBox.rect.height);
				oSpan.style.width = oSpan.offsetWidth - (iGap - 200) + "px";
				oSpan.style.overflow = "hidden";
			}
			else
				oMsgBox.setSize(oMsgBox.rect.width + iGap, oMsgBox.rect.height);
		}
		
		while(true)
		{
			iCurIndex = sMsg.indexOf("<br />", iCurIndex);
			if(iCurIndex == -1)
				break;
			iCurIndex += 4;
			iLines++;
		}
		
		for(iCounter = 0; iCounter < 6; iCounter++)
		{
			if(oSpan.offsetHeight > iLines * 20)
			{
				oMsgBox.setSize(oMsgBox.rect.width + 40, oMsgBox.rect.height);
				if(oMsgBox.rect.left >= 20)
					oMsgBox.setPosition(oMsgBox.rect.left - 20, oMsgBox.rect.top);
			}
			else
				break;
		}
		
		if(oSpan.offsetHeight > 40)
			oMsgBox.setSize(oMsgBox.rect.width, oMsgBox.rect.height + oSpan.offsetHeight - 40);
	}
	
	
	//==========================================================
	//
	//	Function	:	messageBoxOk
	//
	//	Summary		:	Click OK button of a message box
	//
	//==========================================================
	function messageBoxOk(oButton)
	{
		oButton.parent.returnValue = jaswin.MSGBOX_RETURN_OKYES;
		oButton.parent.close();
	}
	
	
	//==========================================================
	//
	//	Function	:	messageBoxCancel
	//
	//	Summary		:	Click cancel button of a message box
	//
	//==========================================================
	function messageBoxCancel(oButton)
	{
		oButton.parent.close();
	}
	
}(window.jaswin = window.jaswin || {}));


//==========================================================
//
//	Function	:	initJaswin
//
//	Summary		:	Initialize jaswin(Call on load html)
//
//==========================================================
function initJaswin()
{
	jaswin.getBrowserName();
	if(jaswin.BROWSER == "")
	{
		alert("This browser is not supported.");
		return;
	}
	
	global = new jaswin.Global();
	
	// Cancel default context menu
	document.oncontextmenu = function () {return false;}
}
